import os, sys, sysconfig, shutil, subprocess, re, keyword, platform, warnings
import pybind11
from .. import CACHE_PATH


def compile_option(type, option):
    if option is None:
        return ""
    if sys.platform in ["linux", "darwin"]:
        if type == "header_search_path":
            return f"\"-I{option}\""
        elif type == "macro_definition":
            return f"\"-D{option}\""
        else:
            raise Exception(f"Unknown option type {type}")
    elif sys.platform.startswith("win"):
        if type == "header_search_path":
            return f"\"/I{option}\""
        elif type == "macro_definition":
            return f"\"/D{option}\""
        else:
            raise Exception(f"Unknown option type {type}")
    else:
        raise Exception(f"Unknown platform {sys.platform}")


absolute_path = os.path.dirname(os.path.abspath(__file__))

def is_valid_module_name(s):
    return bool(re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', s)) and not keyword.iskeyword(s)


def find_compiler():
    if "TINYRL_COMPILER" in os.environ:
        return [os.environ["TINYRL_COMPILER"]]
    unix_compilers = ["clang++", "g++"]
    windows_compilers = ["cl"]
    compilers = unix_compilers if (sys.platform in ["linux", "darwin"]) else windows_compilers
    compilers = [compiler for compiler in compilers if shutil.which(compiler) is not None]
    raw_platform_compiler = platform.python_compiler()
    if raw_platform_compiler.startswith("MSC"):
        platform_compiler = "cl"
    elif raw_platform_compiler.startswith("GCC"):
        platform_compiler = "g++"
    else:
        platform_compiler = None
    if platform_compiler is None:
        warnings.warn(f"The platform compiler {raw_platform_compiler} is not recognized.")
    else:
        if platform_compiler not in compilers:
            warnings.warn(f"The platform compiler {platform_compiler} (used for compiling Python) is not found (found: {compilers}).")
        else:
            compilers = [platform_compiler]
    assert len(compilers) > 0, "No C++ compiler found. Please install clang, g++ or cl (MSVC)."
    return compilers


def compile(source, module, flags=[], enable_optimization=True, force_recompile=False, verbose=False, **kwargs):
    '''
    Takes a link to a source file and compiles it into a shared object.
    Caches the compilation in /tmp/tinyrl/interface/{module}/
    Returns the path to the shared object.
    '''
    assert(is_valid_module_name(module))
    shared_flag = '-shared' if not sys.platform.startswith('win') else '/LD'
    cpp_std_flag = '-std=c++17' if not sys.platform.startswith('win') else '/std:c++17'
    optimization_flag = ('-O3' if not sys.platform.startswith('win') else '/O2') if enable_optimization else ''
    arch_flag = '-march=native' if not sys.platform.startswith('win') else '/arch:AVX2'
    link_math_flag = '-lm' if not sys.platform.startswith('win') else ''
    fast_math_flag = '-ffast-math' if not sys.platform.startswith('win') else '/fp:fast'
    lto_flag = '' #'-flto' if not sys.platform.startswith('win') else '/GL'
    pic_flag = '-fPIC' if not sys.platform.startswith('win') else '/LD'
    link_stdlib_flag = '-stdlib=libc++' if sys.platform == 'darwin' else ''
    verbose_flag = compile_option("macro_definition", 'TINYRL_VERBOSE') if verbose or "TINYRL_VERBOSE" in os.environ else ''

    pybind_includes = [compile_option("header_search_path", pybind11.get_include())]
    python_include_path = sysconfig.get_paths()['include']
    if sys.platform.startswith('win'):
        os.listdir(python_include_path) # check if the path is accessible on Windows (if not, you might need to uninstall the Windows Store version of python and install the installer version from the Python website)
    python_includes = [compile_option("header_search_path", python_include_path)]
    rl_tools_includes = [compile_option("header_search_path", str(os.path.join(absolute_path, "..", "external", "rl_tools", "include")))]

    if sys.platform in ["linux", "darwin"]:
        extension = "so"
        if platform.python_implementation() != "PyPy":
            link_python_args = ["-L"+sysconfig.get_config_var('LIBDIR'), "-lpython" + sysconfig.get_config_var('VERSION')]
        else:
            link_python_args = ["-L"+sysconfig.get_config_var('LIBDIR'), f"-lpypy{sysconfig.get_config_var('VERSION')}-c"]
            extension = f"pypy{sys.version_info.major}{sys.version_info.minor}-pp{sys.pypy_version_info.major}{sys.pypy_version_info.minor}-{sys.platform}.so"
    elif sys.platform.startswith('win'):
        extension = "pyd"
        link_python_args = [f"/link \"/LIBPATH:"+os.path.join(sys.base_exec_prefix, "libs")+"\""]
    
    output_dir = f"{CACHE_PATH}/build/{module}"
    os.makedirs(output_dir, exist_ok=True)
    cmd_path = os.path.join(output_dir, f"cmd.txt")
    output_path = os.path.join(output_dir, f"module.{extension}")

    compilers = find_compiler()

    cmds = [
        [
            compiler,
            shared_flag, 
            pic_flag,
            cpp_std_flag,
            optimization_flag,
            arch_flag,
            fast_math_flag,
            verbose_flag,
            lto_flag,
            *pybind_includes,
            *python_includes,
            *rl_tools_includes,
            f"\"{source}\"",
            *flags,
            link_math_flag,
            link_stdlib_flag,
            *link_python_args,
            *(["-o", output_path] if not sys.platform.startswith('win') else [f"\"/OUT:{output_path}\""]),
        ]
        for compiler in compilers
    ]
    command_strings = [" ".join(cmd) for cmd in cmds]
    old_command_string = None
    if os.path.exists(cmd_path):
        with open(cmd_path, "r") as f:
            old_command_string = f.read()
    if old_command_string is None or (not old_command_string in command_strings) or not os.path.exists(output_path) or force_recompile or "TINYRL_FORCE_RECOMPILE" in os.environ:
        for compiler, cmd, command_string in zip(compilers, cmds, command_strings):
            print(f"Compiling the TinyRL interface...", flush=True)
            verbose_actual = verbose or "TINYRL_FORCE_COMPILE_VERBOSE" in os.environ
            run_kwargs = {"cwd": output_dir} if sys.platform.startswith('win') else {}
            run_kwargs = {**run_kwargs, **({} if verbose_actual else {"capture_output": True, "text": True})}
            print(f"Command: {command_string}", flush=True) if verbose_actual else None
            result = subprocess.run(command_string, check=False, shell=True, **run_kwargs)
            if result.returncode != 0:
                print("Command: ")
                print(command_string)
                if not verbose_actual:
                    print("_______OUTPUT_______")
                    print(result.stdout)
                    print(result.stderr)
                raise Exception(f"Failed to compile the TinyRL interface using {compiler}.")
            with open(cmd_path, "w") as f:
                f.write(command_string)
            print(f"Finished compiling the TinyRL interface.", flush=True)
            break
    else:
        print("Using cached TinyRL interface.", flush=True)
    return output_path


from .accelerate import acceleration_flags
from .load_module import load_module
from .compile import compile, compile_option
import os
absolute_path = os.path.dirname(os.path.abspath(__file__))


def get_time_limit(env):
    max_episode_steps = None
    try:
        max_episode_steps = env.spec.max_episode_steps
    except:
        pass
    return max_episode_steps


import re

def is_valid_identifier(s):
    pattern = r'^[_a-zA-Z][_a-zA-Z0-9]*$'
    return bool(re.match(pattern, s))

def compile_training(module_name, env_factory, flags, EPISODE_STEP_LIMIT=None, verbose=False, force_blas=False, enable_blas=True, force_recompile=False, enable_evaluation=False, evaluation_interval=None, num_evaluation_episodes=10, disable_tensorboard=True): #either gym env or custom env spec
    assert is_valid_identifier(module_name), "Module name is not a valid C identifier"
    assert enable_blas or not force_blas, "Cannot force BLAS acceleration without enabling BLAS"
    use_python_environment = type(env_factory) != dict
    if use_python_environment:
        example_env = env_factory()
        EPISODE_STEP_LIMIT = get_time_limit(example_env) if EPISODE_STEP_LIMIT is None else EPISODE_STEP_LIMIT
    
    if EPISODE_STEP_LIMIT is not None:
        flags += [compile_option("macro_definition", f'TINYRL_EPISODE_STEP_LIMIT={EPISODE_STEP_LIMIT}')]
        
    custom_environment_header_search_path = None if use_python_environment else env_factory["path"]
    custom_environment_flag = compile_option("header_search_path", custom_environment_header_search_path)
    use_python_environment_flag = compile_option("macro_definition", "TINYRL_USE_PYTHON_ENVIRONMENT") if use_python_environment else ''
    observation_dim_flag = compile_option("macro_definition", f'TINYRL_OBSERVATION_DIM={example_env.observation_space.shape[0]}') if use_python_environment else ''
    action_dim_flag = compile_option("macro_definition", f'TINYRL_ACTION_DIM={example_env.action_space.shape[0]}') if use_python_environment else ''
    evaluation_flags = [compile_option("macro_definition", 'TINYRL_ENABLE_EVALUATION') if enable_evaluation else '']
    evaluation_flags += [compile_option("macro_definition", f'TINYRL_EVALUATION_INTERVAL={evaluation_interval}') if evaluation_interval else '']
    evaluation_flags += [compile_option("macro_definition", f'TINYRL_NUM_EVALUATION_EPISODES={num_evaluation_episodes}') if num_evaluation_episodes else '']
    disable_tensorboard = compile_option("macro_definition", 'RLTOOLS_DISABLE_TENSORBOARD') if disable_tensorboard else ''
    module_flag = compile_option("macro_definition", f'TINYRL_MODULE_NAME={module_name}')
    flags += [use_python_environment_flag, custom_environment_flag, observation_dim_flag, action_dim_flag, *evaluation_flags, module_flag]
    if enable_blas:
        flags += acceleration_flags(module_name)
        flags += [compile_option("macro_definition", f'TINYRL_FORCE_BLAS={"true" if force_blas else "false"}') if force_blas else '']

    source = os.path.join(absolute_path, '../interface/training/training.cpp')
    output_path = compile(source, module_name, flags, verbose=verbose, force_recompile=force_recompile) #, **compile_time_parameters)

    module = load_module(module_name, output_path)
    if use_python_environment:
        module.set_environment_factory(env_factory)
    return module


import os
from .render import render, sanitize_values
from .training import compile_training
from .compile import compile_option

from .. import CACHE_PATH


absolute_path = os.path.dirname(os.path.abspath(__file__))

def SAC(env_factory, # can be either a lambda that creates a new Gym-like environment, or a dict with a specification of a C++ environment: {"path": "path/to/environment", "action_dim": xx, "observation_dim": yy}
    verbose=False,
    force_recompile=False,
    enable_evaluation=True,
    evaluation_interval=1000,
    num_evaluation_episodes=10,
    interface_name="default", # this is the namespace used for the compilation of the TinyRL interface (in a temporary directory) and should be unique if run in parallel. We don't choose a random uuid because it would invalidate the cache and require a re-compilation every time
    # Compile-time parameters:
    # Same set of parameters as: rl::algorithms::td3::DefaultParameters
    GAMMA = 0.99,
    ALPHA = 0.5,
    ACTOR_BATCH_SIZE = 100, #32,
    CRITIC_BATCH_SIZE = 100, #32,
    CRITIC_TRAINING_INTERVAL = 1,
    ACTOR_TRAINING_INTERVAL = 1,
    CRITIC_TARGET_UPDATE_INTERVAL = 1,
    ACTOR_POLYAK = 1.0 - 0.005,
    CRITIC_POLYAK = 1.0 - 0.005,
    IGNORE_TERMINATION = False,
    TARGET_ENTROPY = None,
    ADAPTIVE_ALPHA = True,
    ACTION_LOG_STD_LOWER_BOUND=-20,
    ACTION_LOG_STD_UPPER_BOUND=2,
    # Same set of parameters as rl::algorithms::sac::loop::core::DefaultParameters
    N_ENVIRONMENTS = 1,
    N_WARMUP_STEPS = None,
    STEP_LIMIT = 10000,
    REPLAY_BUFFER_CAP = None,
    EPISODE_STEP_LIMIT = None,
    ACTOR_HIDDEN_DIM = 64,
    ACTOR_NUM_LAYERS = 3,
    ACTOR_ACTIVATION_FUNCTION = "RELU",
    CRITIC_HIDDEN_DIM = 64,
    CRITIC_NUM_LAYERS = 3,
    CRITIC_ACTIVATION_FUNCTION = "RELU",
    COLLECT_EPISODE_STATS = True,
    EPISODE_STATS_BUFFER_SIZE = 1000,
    # optimizer
    OPTIMIZER_ALPHA=1e-3,
    OPTIMIZER_BETA_1=0.9,
    OPTIMIZER_BETA_2=0.999,
    OPTIMIZER_EPSILON=1e-7,
    **kwargs
    ):
    verbose = verbose or "TINYRL_VERBOSE" in os.environ


    # The action dim is needed to set the default target entropy
    use_python_environment = type(env_factory) != dict
    if use_python_environment:
        example_env = env_factory()
        ACTION_DIM = example_env.action_space.shape[0]
    else:
        ACTION_DIM = env_factory["action_dim"]

    TARGET_ENTROPY = TARGET_ENTROPY if TARGET_ENTROPY is not None else -ACTION_DIM
    REPLAY_BUFFER_CAP = REPLAY_BUFFER_CAP if REPLAY_BUFFER_CAP is not None else STEP_LIMIT
    N_WARMUP_STEPS = N_WARMUP_STEPS if N_WARMUP_STEPS is not None else max(ACTOR_BATCH_SIZE, CRITIC_BATCH_SIZE)

    compile_time_parameters = sanitize_values(locals())

    module_name = f'tinyrl_sac_{interface_name}'

    config_template = os.path.join(absolute_path, '../interface/algorithms/sac/template.h')

    print('TinyRL Cache Path: ', CACHE_PATH, flush=True) if verbose else None
    render_output_directory = os.path.join(CACHE_PATH, 'template', module_name)
    
    os.makedirs(render_output_directory, exist_ok=True)
    render_output_path = os.path.join(render_output_directory, 'loop_core_config.h')
    new_config = render(config_template, render_output_path, **compile_time_parameters)
    if new_config:
        print('New SAC config detected, forcing recompilation...')
    
    loop_core_config_search_path_flag = compile_option("header_search_path", render_output_directory)
    loop_core_config_flag = compile_option("macro_definition", "TINYRL_USE_LOOP_CORE_CONFIG")
    flags = [loop_core_config_search_path_flag, loop_core_config_flag]
    return compile_training(module_name, env_factory, flags, verbose=verbose, force_recompile=(force_recompile or new_config), enable_evaluation=enable_evaluation, evaluation_interval=evaluation_interval, num_evaluation_episodes=num_evaluation_episodes, **kwargs)



from string import Template
import os

def sanitize_values(kwargs):
    output = {}
    for key, value in kwargs.items():
        if key.isupper():
            if isinstance(value, str):
                output[key] = value
            elif isinstance(value, bool):
                output[key] = "true" if value else "false"
            else:
                output[key] = str(value)
    return output

def render(template_path, output_path, **kwargs):
    '''
    Returns True if a new file was written or if the file was modified.
    '''
    with open(template_path) as f:
        template = Template(f.read())
    rendered = template.substitute(kwargs)
    if os.path.exists(output_path):
        with open(output_path, "r") as f:
            if f.read() == rendered:
                return False
    with open(output_path, "w") as f:
        f.write(rendered)
    return True



import os, sys, warnings
force_mkl = "TINYRL_FORCE_MKL" in os.environ

def link_mkl():
    flags = []
    mkl_found = False
    if sys.platform == "linux":
        if "torch" in sys.modules:
            warning = "PyTorch is imported which is known to cause issues when loading TinyRL with MKL. Set environment variable TINYRL_DISABLE_MKL to disable MKL or disable this warning by TINYRL_IGNORE_TORCH_WARNING."
            if "TINYRL_IGNORE_TORCH_WARNING" not in os.environ:
                raise RuntimeError(warning)
            else:
                warnings.warn(warning)
        from importlib.metadata import files, version, PackageNotFoundError
        try:
            mkl_version = version("mkl")
            mkl_include_version = version("mkl-include")
            mkl_include_files = files("mkl-include")
            mkl_main_header_index = [os.path.basename(str(f)) == "mkl.h" for f in mkl_include_files].index(True)
            mkl_main_header_path = mkl_include_files[mkl_main_header_index]
            mkl_include_path = os.path.dirname(mkl_main_header_path.locate())
            print(f"MKL found. Version: {mkl_version} (include: {mkl_include_version}, path {mkl_include_path})")

            mkl_files = files("mkl")
            # create version symlinks for the MKL libraries (as they are not included in the pypi mkl package)
            required_libraries = ["libmkl_intel_ilp64", "libmkl_intel_thread", "libmkl_core"]
            required_libraries_paths = []
            for lib in required_libraries:
                found = False
                for f in mkl_files:
                    if lib in str(f):
                        required_libraries_paths.append(f)
                        found = True
                        break
                if not found:
                    print(f"MKL library {lib} not found.")
                    required_libraries_paths = None
            required_libraries_paths_absolute = [str(p.locate()) for p in required_libraries_paths]
            required_libraries_search_paths = [str(os.path.dirname(p)) for p in required_libraries_paths_absolute]
            flags += [
                "-m64",
                "-Wl,--no-as-needed",
                *required_libraries_paths_absolute,
                "-liomp5",
                "-lpthread",
                "-lm",
                "-ldl",
                "-I" + mkl_include_path,
                *[f"-L{p}" for p in required_libraries_search_paths],
                "-Wl,--rpath," + ":".join(required_libraries_search_paths),
            ]
            flags += ["-I" + os.path.join(sys.prefix + "/include")]
            flags += ["-DRL_TOOLS_BACKEND_ENABLE_MKL"]
            flags += ["-DRL_TOOLS_DISABLE_UNALIGNED_MEMORY_ALLOCATIONS"]
            mkl_found = True
        except PackageNotFoundError:
            assert(not force_mkl)
            print("MKL is not installed. To use MKL please install `mkl` and `mkl-include`")
    return mkl_found, flags


import sys
import importlib
import importlib.util
import importlib.abc

def load_module(module_name, path, force_reload=True):
    spec = importlib.util.spec_from_file_location(module_name, path)
    assert spec is not None
    module = importlib.util.module_from_spec(spec)
    assert isinstance(spec.loader, importlib.abc.Loader)
    spec.loader.exec_module(module)
    if module_name not in sys.modules:
        sys.modules[module_name] = module
    return module


import sys, os
from .link_accelerate import link_accelerate
from .link_mkl import link_mkl
from .link_blas import link_blas

disable_mkl = "TINYRL_DISABLE_MKL" in os.environ

def acceleration_flags(module):
    flags = []
    success = False
    if sys.platform == "darwin":
        success, accelerate_flags = link_accelerate()
        if success:
            flags += accelerate_flags
    elif sys.platform == "linux":
        if not disable_mkl:
            success, mkl_flags = link_mkl()
            if success:
                flags += mkl_flags
        if not success:
            success, blas_flags = link_blas(module)
            flags += blas_flags
    return flags


import os, sys, subprocess

from .. import CACHE_PATH

from .compile import find_compiler
force_mkl = "TINYRL_FORCE_MKL" in os.environ

def link_blas(module):
    flags = []
    blas_found = False
    if sys.platform == "linux":
        compiler = find_compiler()[0]
        output_dir = f"{CACHE_PATH}/blas/{module}"
        os.makedirs(output_dir, exist_ok=True)
        output_test_file_path = os.path.join(output_dir, "test_blas.cpp")
        output_test_output_path = os.path.join(output_dir, "a.out")
        with open(output_test_file_path, "w") as f:
            f.write("int main() { return 0; }")
        
        result = subprocess.run(f"{compiler} -lblas {output_test_file_path} -o {output_test_output_path}", check=False, shell=True)
        if result.returncode == 0:
            print(f"BLAS found.")
            flags += ["-lblas"]
            flags += ["-DRL_TOOLS_BACKEND_ENABLE_OPENBLAS"]
            flags += ["-DRL_TOOLS_DISABLE_UNALIGNED_MEMORY_ALLOCATIONS"]
            blas_found = True
    return blas_found, flags


import os, math
import inspect
from .render import render, sanitize_values
from .training import compile_training
from .compile import compile_option

from .. import CACHE_PATH


absolute_path = os.path.dirname(os.path.abspath(__file__))

def PPO(env_factory, # can be either a lambda that creates a new Gym-like environment, or a dict with a specification of a C++ environment: {"path": "path/to/environment", "action_dim": xx, "observation_dim": yy}
    verbose=False,
    force_recompile=False,
    enable_evaluation=True,
    evaluation_interval=None,
    num_evaluation_episodes=10,
    interface_name="default", # this is the namespace used for the compilation of the TinyRL interface (in a temporary directory) and should be unique if run in parallel. We don't choose a random uuid because it would invalidate the cache and require a re-compilation every time
    # Compile-time parameters:
    # Same set of parameters as: rl::algorithms::ppo::DefaultParameters
    GAMMA = 0.99,
    LAMBDA = 0.95,
    EPSILON_CLIP = 0.2,
    INITIAL_ACTION_STD = 0.5,
    LEARN_ACTION_STD = True,
    ACTION_ENTROPY_COEFFICIENT = 0.01,
    ADVANTAGE_EPSILON = 1e-8,
    NORMALIZE_ADVANTAGE = True,
    ADAPTIVE_LEARNING_RATE = False,
    ADAPTIVE_LEARNING_RATE_POLICY_KL_THRESHOLD = 0.008,
    POLICY_KL_EPSILON = 1e-5,
    ADAPTIVE_LEARNING_RATE_DECAY = 1/1.5,
    ADAPTIVE_LEARNING_RATE_MIN = 1e-6,
    ADAPTIVE_LEARNING_RATE_MAX = 1e-2,
    NORMALIZE_OBSERVATIONS = False,
    N_WARMUP_STEPS_CRITIC = 0,
    N_WARMUP_STEPS_ACTOR = 0,
    N_EPOCHS = 10,
    IGNORE_TERMINATION = False, # ignoring the termination flag is useful for training on environments with negative rewards, where the agent would try to terminate the episode as soon as possible otherwise
    # Same set of parameters as rl::algorithms::ppo::loop::core::DefaultParameters
    STEP_LIMIT = None,
    TOTAL_STEP_LIMIT = 100000, # This is environment step limit => translated into the closets ppo step limit (smaller or equal in total environment steps)
    ACTOR_HIDDEN_DIM = 64,
    ACTOR_NUM_LAYERS = 3,
    ACTOR_ACTIVATION_FUNCTION = "RELU",
    CRITIC_HIDDEN_DIM = 64,
    CRITIC_NUM_LAYERS = 3,
    CRITIC_ACTIVATION_FUNCTION = "RELU",
    EPISODE_STEP_LIMIT = 1000,
    N_ENVIRONMENTS = 64,
    ON_POLICY_RUNNER_STEPS_PER_ENV = 64,
    BATCH_SIZE = 64,
    # optimizer
    OPTIMIZER_ALPHA=1e-3,
    OPTIMIZER_BETA_1=0.9,
    OPTIMIZER_BETA_2=0.999,
    OPTIMIZER_EPSILON=1e-7,
    **kwargs
    ):
    assert STEP_LIMIT is not None or TOTAL_STEP_LIMIT is not None, "Either STEP_LIMIT or TOTAL_STEP_LIMIT must be set"
    evaluation_interval = evaluation_interval if evaluation_interval is not None else 10
    verbose = verbose or "TINYRL_VERBOSE" in os.environ
    if STEP_LIMIT is None:
        STEP_LIMIT = math.floor(TOTAL_STEP_LIMIT/(ON_POLICY_RUNNER_STEPS_PER_ENV * N_ENVIRONMENTS))

    compile_time_parameters = sanitize_values({k:v for k, v in locals().items() if k in inspect.signature(PPO).parameters.keys()})

    module_name = f'tinyrl_ppo_{interface_name}'

    config_template = os.path.join(absolute_path, '../interface/algorithms/ppo/template.h')

    print('TinyRL Cache Path: ', CACHE_PATH) if verbose else None
    render_output_directory = os.path.join(CACHE_PATH, 'template', module_name)
    
    os.makedirs(render_output_directory, exist_ok=True)
    render_output_path = os.path.join(render_output_directory, 'loop_core_config.h')
    new_config = render(config_template, render_output_path, **compile_time_parameters)
    if new_config:
        print('New PPO config detected, forcing recompilation...')
    
    loop_core_config_search_path_flag = compile_option("header_search_path", render_output_directory)
    loop_core_config_flag = compile_option("macro_definition", "TINYRL_USE_LOOP_CORE_CONFIG")
    flags = [loop_core_config_search_path_flag, loop_core_config_flag]
    return compile_training(module_name, env_factory, flags, verbose=verbose, force_recompile=(force_recompile or new_config), enable_evaluation=enable_evaluation, evaluation_interval=evaluation_interval, num_evaluation_episodes=num_evaluation_episodes, **kwargs)



import os, hashlib
from .. import CACHE_PATH
from .load_module import load_module
from .compile import compile
from .accelerate import acceleration_flags

def load_checkpoint_from_path(checkpoint_path, interface_name="default", force_recompile=False, verbose=False):
    with open(checkpoint_path, "r") as f:
        checkpoint = f.read()
    return load_checkpoint(checkpoint, interface_name="default", force_recompile=force_recompile, verbose=verbose)

def load_checkpoint(checkpoint, interface_name="default", force_recompile=False, verbose=False):
    module_name = f"load_checkpoint_{interface_name}"
    output_directory = os.path.join(CACHE_PATH, "checkpoint", module_name)
    
    os.makedirs(output_directory, exist_ok=True)
    output_path = os.path.join(output_directory, 'checkpoint.h')
    with open(output_path, "w") as f:
        f.write(checkpoint)

    module_flag = f'-DTINYRL_MODULE_NAME={module_name}'
    header_search_path_flag = f'-I{output_directory}'

    flags = [module_flag, header_search_path_flag]

    flags += acceleration_flags(module_name)

    absolute_path = os.path.dirname(os.path.abspath(__file__))
    source = os.path.join(absolute_path, '../interface/inference/inference.cpp')
    output_path = compile(source, module_name, flags, force_recompile=force_recompile, verbose=verbose)
    module = load_module(module_name, output_path)
    return module

